import torch
from datasets import load_dataset, Dataset as HfDataset
from torch.utils.data import Dataset
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from typing import Union, TypeAlias, Literal
from ..training.dataset import BaseDataset

class MaskedLMDataset(BaseDataset):
    def __init__(
            self,
            texts: Union[list[str], HfDataset],
            tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
            max_seq_len: int = 1024,
            mask_prob: float = 0.15,
            hf_field: str = 'text',
            *args,
            **kwargs
    ):
        super(MaskedLMDataset, self).__init__(texts, tokenizer, max_seq_len, hf_field, *args, **kwargs)
        self.mask_prob = mask_prob

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        inputs = self.get_tokenized_text(idx)

        input_ids = inputs['input_ids'][0]
        if self.is_pre_tokenized:
            input_ids = input_ids.clone()
        attention_mask = inputs['attention_mask'][0]
        labels = input_ids.clone()

        # Create masked indices
        masked_indices = torch.bernoulli(
            torch.full(labels.shape, self.mask_prob)
        ).bool() & attention_mask.bool()

        # Apply mask
        labels[~masked_indices] = -100
        input_ids[masked_indices] = self.tokenizer.mask_token_id

        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'labels': labels
        }


class AutoregressiveLMDataset(BaseDataset):
    def __init__(
            self,
            texts: Union[list[str], HfDataset],
            tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
            max_seq_len: int = 1024,
            hf_field: str = 'text',
            *args,
            **kwargs
    ):
        super(AutoregressiveLMDataset, self).__init__(texts, tokenizer, max_seq_len, hf_field, *args, **kwargs)

    def __getitem__(self, idx: int) -> dict[str, torch.Tensor]:
        inputs = self.get_tokenized_text(idx)

        input_ids = inputs['input_ids'][0]
        attention_mask = inputs['attention_mask'][0]
        targets = input_ids.clone()

        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'targets': targets
        }

ItemFields: TypeAlias = Literal['input_ids', 'attention_mask', 'targets']
SftDataItem: TypeAlias = dict[ItemFields, torch.Tensor]

class DecoderOnlySftDataset(Dataset):
    def __init__(
            self,
            episodes: Union[list[dict], HfDataset],
            tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
            max_seq_len: int = 1024,
            query_field: str = 'query',
            answer_field: str = 'answer',
            interactions_field: str = 'interactions',
            query_token: str = '[Q]',
            answer_token: str = '[A]',
            bos_token: str = '[BOS]',
            eos_token: str = '[EOS]',
            ignore_index: int = -100,
            **kwargs,
    ):
        super(DecoderOnlySftDataset, self).__init__(**kwargs)
        self.episodes = episodes
        self.tokenizer = tokenizer
        self.max_seq_len = max_seq_len
        self.query_field = query_field
        self.answer_field = answer_field
        self.interactions_field = interactions_field
        self.query_token = query_token
        self.answer_token = answer_token
        self.bos_token = bos_token
        self.eos_token = eos_token
        self.ignore_index = ignore_index
        self.is_pre_tokenized = False
        self.is_list = isinstance(self.episodes, list)
        self.inputs = []

    def _build_full_conversation_text(self, item: dict) -> str:
        full_text = f"{self.bos_token}{self.query_token}{item[self.query_field]}{self.answer_token}{item[self.answer_field]}"

        for inter in item[self.interactions_field]:
            full_text += f"{self.query_token}{inter[self.query_field]}{self.answer_token}{inter[self.answer_field]}"

        full_text += self.eos_token
        return full_text

    def _create_masked_labels(self, input_ids: torch.Tensor) -> torch.Tensor:
        labels = input_ids.clone()

        query_token_id = self.tokenizer.convert_tokens_to_ids(self.query_token)
        answer_token_id = self.tokenizer.convert_tokens_to_ids(self.answer_token)
        eos_token_id = self.tokenizer.convert_tokens_to_ids(self.eos_token)

        in_answer = False
        for i in range(len(input_ids)):
            token = input_ids[i].item()
            if token == answer_token_id:
                in_answer = True
            elif token in (query_token_id, eos_token_id):
                in_answer = False

            if not in_answer or token == answer_token_id:
                labels[i] = self.ignore_index

        return labels

    def get_tokenized_item(self, idx: int, episode: dict = None) -> SftDataItem:
        if self.is_pre_tokenized:
            return self.inputs[idx]
        else:
            item = self.episodes[idx] if episode is None else episode
            full_text = self._build_full_conversation_text(item)

            enc = self.tokenizer(
                full_text,
                max_length=self.max_seq_len,
                padding='max_length',
                truncation=True,
                return_tensors='pt',
                add_special_tokens=False
            )

            input_ids = enc['input_ids'][0]
            attention_mask = enc['attention_mask'][0]
            targets = self._create_masked_labels(input_ids)

            return {
                'input_ids': input_ids,
                'attention_mask': attention_mask,
                'targets': targets,
            }

    def __getitem__(self, idx: int) -> SftDataItem:
        return self.get_tokenized_item(idx)

    def __len__(self) -> int:
        return len(self.inputs if self.is_pre_tokenized else self.episodes)

    def get_subset(self, size: float, from_start: bool = False, **kwargs) -> "DecoderOnlySftDataset":
        split_point = int(
            len(self.inputs if self.is_pre_tokenized else self.episodes) * ((1 - size) if not from_start else size))
        if not isinstance(self.episodes, list):
            subset = self.episodes.select(
                range(split_point, len(self.episodes)) if not from_start else range(split_point))
            self.episodes = self.episodes.select(
                range(split_point) if not from_start else range(split_point, len(self.episodes)))
        else:
            subset = self.episodes[split_point:-1] if not from_start else self.episodes[0:split_point]
            self.episodes = self.episodes[0:split_point] if not from_start else self.episodes[split_point:-1]
        return self.__class__(subset, tokenizer=self.tokenizer, query_field=self.query_field,
                              answer_field=self.answer_field, interactions_field=self.interactions_field,
                              max_seq_len=self.max_seq_len, ignore_index=self.ignore_index, **kwargs)

    def pre_tokenize(self, verbose: bool = False, log_interval: int = 10_000, keep_order: bool = False):
        """
        Pre-tokenizes all the items in the dataset, for faster training. Training with pre-tokenized
        dataset could be even 2x faster.

        !! This method has extremely high memory usage, when used with HuggingFace datasets,
        because of converting it to list. Additionally, for the most optimal performance,
        pre-tokenized items are in reversed order - it shouldn't matter for training, as
        items are shuffled then by DataLoader, but you should keep that in mind in case
        of reproducibility.

        Args:
            verbose (bool): Should display logs (default: False)
            log_interval (int): Display logs every log_interval iterations (default: 10_000)
            keep_order (bool): Keep tokenized items in the same order - by default they are reversed for faster processing (default: False)
        """
        if not self.is_pre_tokenized:
            num_episodes = len(self.episodes)
            eps = self.episodes if self.is_list else self.episodes.to_list()
            del self.episodes
            self.episodes = None
            for index in range(num_episodes):
                self.inputs.append(self.get_tokenized_item(index, episode=eps.pop() if not keep_order else eps[index]))
                if verbose and index % log_interval == 0:
                    print(f'Processed {index + 1}/{num_episodes}')
            del eps
            self.is_pre_tokenized = True

    @classmethod
    def from_hf_hub(
            cls,
            dataset_id: str,
            mrl_subset: str,
            tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
            split: str = 'train',
            query_field: str = 'query',
            answer_field: str = 'answer',
            interactions_field: str = 'interactions',
            load_kwargs: dict = None,
            max_seq_len: int = 1024,
            ignore_index: int = -100,
            **kwargs
    ):
        """
        Load dataset from HuggingFace Hub and convert it to RxNN training dataset.

        One of the `tokenizer` or `tokenizer_hub_id` args must be provided. If both are provided, `tokenizer` will be used.

        Args:
            dataset_id (str): Hub dataset repository name
            mrl_subset (str): Dataset subset
            tokenizer (Union[PreTrainedTokenizer, PreTrainedTokenizerFast]): Tokenizer
            split (str): Dataset split (default: "train")
            query_field (str): Query field (default: "query")
            answer_field (str): Answer field (default: "answer")
            interactions_field (str): Interactions field (default: "interactions")
            load_kwargs (dict): Additional args for HuggingFace API load_dataset function
            max_seq_len (int): Maximum sequence length (default: 1024)
            ignore_index (int): Ignore index for labels masking (default: -100)
            **kwargs: Additional args for RxNN Dataset class
        """
        if load_kwargs is None:
            load_kwargs = {}

        hf_dataset = load_dataset(dataset_id, mrl_subset, split=split, **load_kwargs)

        return cls(hf_dataset, tokenizer, query_field=query_field, answer_field=answer_field,
                   interactions_field=interactions_field, max_seq_len=max_seq_len,
                   ignore_index=ignore_index, **kwargs)



import os, traceback, shutil
import numpy as np
import torch
import torch.nn as nn
from typing import Union, Optional, TypeAlias, Literal
from torch.nn.parallel import DistributedDataParallel
from huggingface_hub import PyTorchModelHubMixin
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from ..utils import human_format


class TrainerCallback:
    def on_epoch_start(self, model: nn.Module, epoch: int) -> None:
        pass

    def on_epoch_end(self, model: nn.Module, epoch: int) -> Union[bool, None]:
        pass

    def on_batch_start(self, model: nn.Module, batch_idx: int, batch: dict[str, torch.Tensor]) -> None:
        pass

    def on_batch_end(self, model: nn.Module, batch_idx: int, loss: float, batch: dict[str, torch.Tensor]) -> \
            Union[
                bool, None]:
        pass

    def on_training_end(self, model: nn.Module) -> None:
        pass

    def on_validation_end(self, model: nn.Module, epoch: int, val_loss: float, val_metrics: dict) -> Union[
        bool, None]:
        pass


class PrintLossCallback(TrainerCallback):
    def __init__(self, batch_log_interval: int = 100, joint_mode: bool = False, batches_per_epoch: int = None):
        self.epoch_means = []
        self.epoch_losses = []
        self.batch_group_losses = []
        self.batch_log_interval = batch_log_interval
        self.joint_mode = joint_mode
        self.batches_per_epoch = batches_per_epoch

    def on_batch_start(self, model: nn.Module, batch_idx: int, batch: dict[str, torch.Tensor]) -> None:
        pass

    def on_batch_end(self, model: nn.Module, batch_idx: int, loss: int,
                     batch: dict[str, torch.Tensor]) -> None:
        self.batch_group_losses.append(loss)
        self.epoch_losses.append(loss)

        if batch_idx != 0 and batch_idx % self.batch_log_interval == 0:
            batch_group_mean = np.stack(self.batch_group_losses).mean()
            self.batch_group_losses = []
            if self.batches_per_epoch is not None:
                print(
                    f'Batch {batch_idx} / {self.batches_per_epoch} - loss: {loss}, last {self.batch_log_interval} batches mean loss: {batch_group_mean:.4f}')
            else:
                print(
                    f'Batch {batch_idx} - loss: {loss}, last {self.batch_log_interval} batches mean loss: {batch_group_mean:.4f}')

    def on_epoch_start(self, model: nn.Module, epoch: int) -> None:
        self.epoch_losses = []
        print(f'Start epoch: {epoch}')

    def on_epoch_end(self, model: nn.Module, epoch: int) -> None:
        epoch_mean = np.stack(self.epoch_losses).mean()
        print(f'Epoch {epoch} - mean loss: {epoch_mean:.4f}')
        self.epoch_means.append(epoch_mean)

    def on_training_end(self, model: nn.Module) -> None:
        print(f'Finished training! All losses:')
        print(self.epoch_means)

    def on_validation_end(self, model: nn.Module, epoch: int, val_loss: float, val_metrics: dict) -> None:
        if self.joint_mode:
            print(f"Epoch {epoch} - encoder loss: {val_metrics['loss']['encoder']:.4f}")
            print(f"Epoch {epoch} - decoder loss: {val_metrics['loss']['decoder']:.4f}")
        print(f"Epoch {epoch} - validation Loss: {val_loss:.4f}")


class PrintAccuracyCallback(TrainerCallback):
    def __init__(self, joint_mode: bool = False):
        self.joint_mode = joint_mode

    def on_validation_end(self, model: nn.Module, epoch: int, val_loss: float, val_metrics: dict) -> None:
        if self.joint_mode:
            print(f"Epoch {epoch} - encoder node accuracy: {val_metrics['accuracy']['node_encoder']:.4f}")
            print(f"Epoch {epoch} - decoder node accuracy: {val_metrics['accuracy']['node_decoder']:.4f}")
            print(f"Epoch {epoch} - encoder accuracy: {val_metrics['accuracy']['encoder']:.4f}")
            print(f"Epoch {epoch} - decoder accuracy: {val_metrics['accuracy']['decoder']:.4f}")
        else:
            print(f"Epoch {epoch} - node accuracy: {val_metrics['node_accuracy']:.4f}")
            print(f"Epoch {epoch} - accuracy: {val_metrics['accuracy']:.4f}")


class PrintMemoryAttentionMetricsCallback(TrainerCallback):
    def on_validation_end(self, model: nn.Module, epoch: int, val_loss: float, val_metrics: dict) -> None:
            print(f"Epoch {epoch} - cosine sim: {val_metrics['cosine_sim']:.4f}")
            print(f"Epoch {epoch} - STM RMSE diff: {val_metrics['rmse_diff']:.4f}")


class TokenCounterCallback(TrainerCallback):
    def __init__(self, limit: int, batch_log_interval: int = 100):
        self.total_tokens = 0
        self.limit = limit
        self.batch_log_interval = batch_log_interval

    def on_batch_end(self, model: nn.Module, batch_idx: int, loss: int,
                     batch: dict[str, torch.Tensor]) -> bool:
        attention_mask = batch['attention_mask']
        batch_tokens = attention_mask.sum().item()
        self.total_tokens += batch_tokens
        if batch_idx != 0 and batch_idx % self.batch_log_interval == 0:
            print(f'Total processed tokens: {human_format(self.total_tokens)}')

        should_stop_training = self.total_tokens >= self.limit
        if should_stop_training:
            print(f'Reached a limit of {human_format(self.limit)} processed tokens - stopping training')
        return should_stop_training

    def on_training_end(self, model: nn.Module) -> None:
        print(f'Total training tokens: {human_format(self.total_tokens)}')

    def get_total_tokens(self):
        return self.total_tokens


class ModelSaveCallback(TrainerCallback):
    def __init__(
            self,
            save_dir: str,
            max_keep: int = 3,
            push_to_hub: bool = False,
            hub_model_id: str = None,
            private_repo: bool = False,
            hf_token: str = None,
            push_checkpoint_weights: bool = True,
            final_commit_message: str = None,
            save_checkpoint_after_n_batches: int = None,
            push_batch_checkpoint: bool = False,
            display_exc_trace: bool = False,
            use_ddp: bool = False,
    ):
        self.save_dir = save_dir
        self.max_keep = max_keep
        self.best_loss = float('inf')
        self.ckpt_paths = []
        self.push_to_hub = push_to_hub
        self.hub_model_id = hub_model_id
        self.private_repo = private_repo
        self.hf_token = hf_token
        self.push_checkpoint_weights = push_checkpoint_weights
        self.final_commit_message = final_commit_message
        self.save_checkpoint_after_n_batches = save_checkpoint_after_n_batches
        self.push_batch_checkpoint = push_batch_checkpoint
        self.finished_epochs = 0
        self.display_exc_trace = display_exc_trace
        self.rank = int(os.environ['RANK']) if use_ddp else 0

    def on_batch_end(self, model: nn.Module, batch_idx: int, loss: int, batch: dict[str, torch.Tensor]) -> Union[
        bool, None]:
        if self.rank == 0 and self.save_checkpoint_after_n_batches is not None and batch_idx != 0 and batch_idx % self.save_checkpoint_after_n_batches == 0:
            if isinstance(model, DistributedDataParallel):
                model = next(model.children())
            try:
                if model.save_pretrained is not None:
                    ckpt_path = os.path.join(
                        self.save_dir,
                        'batch_checkpoint'
                    )
                    path_exists = os.path.exists(ckpt_path)
                    if not path_exists:
                        os.makedirs(ckpt_path)
                    model.save_pretrained(save_directory=ckpt_path)
                else:
                    path_exists = os.path.exists(self.save_dir)
                    if not path_exists:
                        os.makedirs(self.save_dir)
                    ckpt_path = os.path.join(
                        self.save_dir,
                        'batch_checkpoint.pt'
                    )
                    os.remove(ckpt_path)
                    torch.save(model.state_dict(), ckpt_path)
            except Exception as e:
                print(f"Error saving batch checkpoint: {str(e)}")
                if self.display_exc_trace:
                    traceback.print_exc()
            try:
                if self.push_to_hub and self.push_batch_checkpoint and model.push_to_hub is not None and self.hub_model_id:
                    model.push_to_hub(
                        repo_id=self.hub_model_id,
                        token=self.hf_token,
                        private=self.private_repo,
                    )
            except Exception as e:
                print(f"Error pushing batch checkpoint: {str(e)}")
                if self.display_exc_trace:
                    traceback.print_exc()

    def on_validation_end(
            self,
            model: Union[nn.Module, PyTorchModelHubMixin],
            epoch: int,
            val_loss: float,
            val_metrics: dict
    ):
        if self.rank == 0:
            self.finished_epochs += 1
            if val_loss < self.best_loss:
                self.best_loss = val_loss
                if isinstance(model, DistributedDataParallel):
                    model = next(model.children())
                try:
                    if model.save_pretrained is not None:
                        ckpt_path = os.path.join(
                            self.save_dir,
                            f'epoch_{epoch}_val_loss_{val_loss:.4f}'
                        )
                        path_exists = os.path.exists(ckpt_path)
                        if not path_exists:
                            os.makedirs(ckpt_path)
                        model.save_pretrained(save_directory=ckpt_path)
                    else:
                        path_exists = os.path.exists(self.save_dir)
                        if not path_exists:
                            os.makedirs(self.save_dir)
                        ckpt_path = os.path.join(
                            self.save_dir,
                            f'epoch_{epoch}_val_loss_{val_loss:.4f}.pt'
                        )
                        torch.save(model.state_dict(), ckpt_path)
                    self.ckpt_paths.append(ckpt_path)

                    # Keep only N best checkpoints
                    if len(self.ckpt_paths) > self.max_keep:
                        oldest_path = self.ckpt_paths.pop(0)
                        if model.save_pretrained is not None:
                            shutil.rmtree(oldest_path)
                        else:
                            os.remove(oldest_path)
                except Exception as e:
                    print(f"Error saving epoch checkpoint: {str(e)}")
                    if self.display_exc_trace:
                        traceback.print_exc()

                try:
                    if self.push_to_hub and self.push_checkpoint_weights and model.push_to_hub is not None and self.hub_model_id:
                        model.push_to_hub(
                            repo_id=self.hub_model_id,
                            commit_message=f'Epoch {epoch} - Val loss {val_loss:.4f}',
                            token=self.hf_token,
                            private=self.private_repo,
                        )
                except Exception as e:
                    print(f"Error pushing epoch checkpoint: {str(e)}")
                    if self.display_exc_trace:
                        traceback.print_exc()

    def on_training_end(self, model: Union[nn.Module, PyTorchModelHubMixin]):
        if self.rank == 0:
            if isinstance(model, DistributedDataParallel):
                model = next(model.children())
            try:
                # Save final model
                if model.save_pretrained is not None:
                    ckpt_path = os.path.join(
                        self.save_dir,
                        'final_model'
                    )
                    model.save_pretrained(save_directory=ckpt_path)
                else:
                    ckpt_path = os.path.join(self.save_dir, 'final_model.pt')
                    torch.save(model.state_dict(), ckpt_path)
                print(f"Final model saved to {ckpt_path}")
            except Exception as e:
                print(f"Error saving final model: {str(e)}")
                if self.display_exc_trace:
                    traceback.print_exc()
            try:
                if self.push_to_hub and model.push_to_hub is not None:
                    model.push_to_hub(
                        repo_id=self.hub_model_id,
                        commit_message=self.final_commit_message or f'Final pre-trained model, after {self.finished_epochs} epochs',
                        token=self.hf_token,
                        private=self.private_repo,
                    )
                print(f"Model uploaded to repo: {self.hub_model_id}")
            except Exception as e:
                print(f"Error pushing final model: {str(e)}")
                if self.display_exc_trace:
                    traceback.print_exc()


ModelToSave: TypeAlias = Literal['decoder', 'encoder', 'head', 'mem_attn']


class JointModelSaveCallback(TrainerCallback):
    def __init__(
            self,
            save_dir: str,
            max_keep: int = 3,
            push_to_hub: bool = False,
            hub_model_decoder: str = None,
            hub_model_encoder: str = None,
            hub_model_head: str = None,
            hub_model_mem_attn: str = None,
            private_repo: bool = False,
            hf_token: str = None,
            push_checkpoint_weights: bool = True,
            final_commit_message: str = None,
            save_checkpoint_after_n_batches: int = None,
            push_batch_checkpoint: bool = False,
            save_models: list[ModelToSave] = ('decoder', 'encoder', 'head', 'mem_attn'),
            display_exc_trace: bool = False,
            use_ddp: bool = False,
    ):
        self.save_dir = save_dir
        self.max_keep = max_keep
        self.best_loss = float('inf')
        self.ckpt_paths = []
        self.push_to_hub = push_to_hub
        self.hub_model_decoder = hub_model_decoder
        self.hub_model_encoder = hub_model_encoder
        self.hub_model_head = hub_model_head
        self.hub_model_mem_attn = hub_model_mem_attn
        self.private_repo = private_repo
        self.hf_token = hf_token
        self.push_checkpoint_weights = push_checkpoint_weights
        self.final_commit_message = final_commit_message
        self.save_checkpoint_after_n_batches = save_checkpoint_after_n_batches
        self.push_batch_checkpoint = push_batch_checkpoint
        self.finished_epochs = 0
        self.save_models = save_models
        self.display_exc_trace = display_exc_trace
        self.rank = int(os.environ['RANK']) if use_ddp else 0

    def _save_batch(self, model: Union[nn.Module, PyTorchModelHubMixin], component: str, hub_id: str = None):
        try:
            if model.save_pretrained is not None:
                ckpt_path = os.path.join(
                    self.save_dir,
                    component,
                    'batch_checkpoint'
                )
                path_exists = os.path.exists(ckpt_path)
                if not path_exists:
                    os.makedirs(ckpt_path)
                model.save_pretrained(save_directory=ckpt_path)
            else:
                comp_path = os.path.join(
                    self.save_dir,
                    component
                )
                path_exists = os.path.exists(comp_path)
                if not path_exists:
                    os.makedirs(comp_path)
                ckpt_path = os.path.join(
                    comp_path,
                    'batch_checkpoint.pt'
                )
                os.remove(ckpt_path)
                torch.save(model.state_dict(), ckpt_path)
        except Exception as e:
            print(f"Error saving batch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()
        try:
            if self.push_to_hub and self.push_batch_checkpoint and model.push_to_hub is not None and hub_id:
                model.push_to_hub(
                    repo_id=hub_id,
                    token=self.hf_token,
                    private=self.private_repo,
                )
        except Exception as e:
            print(f"Error pushing batch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

    def on_batch_end(self, model: nn.Module, batch_idx: int, loss: int, batch: dict[str, torch.Tensor]) -> Union[
        bool, None]:
        if self.rank == 0 and self.save_checkpoint_after_n_batches is not None and batch_idx != 0 and batch_idx % self.save_checkpoint_after_n_batches == 0:
            if isinstance(model, DistributedDataParallel):
                model = next(model.children())

            if 'decoder' in self.save_models:
                batch_size = model.decoder.model.stm.batch_size
                model.decoder.model.stm.single_memory()
                self._save_batch(model.decoder, 'decoder', hub_id=self.hub_model_decoder)
                model.decoder.model.stm.batched_memory(batch_size)

            if 'encoder' in self.save_models:
                self._save_batch(model.encoder, 'encoder', hub_id=self.hub_model_encoder)

            if 'head' in self.save_models:
                self._save_batch(model.mlm_head, 'head', hub_id=self.hub_model_head)

            if 'mem_attn' in self.save_models:
                batch_size = model.memory_attention.model.stm.batch_size
                model.memory_attention.model.stm.single_memory()
                self._save_batch(model.memory_attention, 'mem_attn', hub_id=self.hub_model_mem_attn)
                model.memory_attention.model.stm.batched_memory(batch_size)

    def _save_validation(self, model: Union[nn.Module, PyTorchModelHubMixin], component: str, epoch: int,
                         val_loss: float, hub_id: str = None):
        try:
            if model.save_pretrained is not None:
                ckpt_path = os.path.join(
                    self.save_dir,
                    component,
                    f'epoch_{epoch}_val_loss_{val_loss:.4f}'
                )
                path_exists = os.path.exists(ckpt_path)
                if not path_exists:
                    os.makedirs(ckpt_path)
                model.save_pretrained(save_directory=ckpt_path)
            else:
                comp_path = os.path.join(
                    self.save_dir,
                    component
                )
                path_exists = os.path.exists(comp_path)
                if not path_exists:
                    os.makedirs(comp_path)
                ckpt_path = os.path.join(
                    comp_path,
                    f'epoch_{epoch}_val_loss_{val_loss:.4f}.pt'
                )
                torch.save(model.state_dict(), ckpt_path)
            self.ckpt_paths.append(ckpt_path)

            # Keep only N best checkpoints
            if len(self.ckpt_paths) > self.max_keep:
                oldest_path = self.ckpt_paths.pop(0)
                if model.save_pretrained is not None:
                    shutil.rmtree(oldest_path)
                else:
                    os.remove(oldest_path)
        except Exception as e:
            print(f"Error saving epoch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

        try:
            if self.push_to_hub and self.push_checkpoint_weights and model.push_to_hub is not None and hub_id:
                model.push_to_hub(
                    repo_id=hub_id,
                    commit_message=f'Epoch {epoch} - Val loss {val_loss:.4f}',
                    token=self.hf_token,
                    private=self.private_repo,
                )
        except Exception as e:
            print(f"Error pushing epoch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

    def on_validation_end(
            self,
            model: Union[nn.Module, PyTorchModelHubMixin],
            epoch: int,
            val_loss: float,
            val_metrics: dict
    ):
        if self.rank == 0:
            self.finished_epochs += 1
            if val_loss < self.best_loss:
                self.best_loss = val_loss
                if isinstance(model, DistributedDataParallel):
                    model = next(model.children())

                if 'decoder' in self.save_models:
                    batch_size = model.decoder.model.stm.batch_size
                    model.decoder.model.stm.single_memory()
                    self._save_validation(model.decoder, 'decoder', epoch, val_loss, hub_id=self.hub_model_decoder)
                    model.decoder.model.stm.batched_memory(batch_size)

                if 'encoder' in self.save_models:
                    self._save_validation(model.encoder, 'encoder', epoch, val_loss, hub_id=self.hub_model_encoder)

                if 'head' in self.save_models:
                    self._save_validation(model.mlm_head, 'head', epoch, val_loss, hub_id=self.hub_model_head)

                if 'mem_attn' in self.save_models:
                    batch_size = model.memory_attention.model.stm.batch_size
                    model.memory_attention.model.stm.single_memory()
                    self._save_validation(model.memory_attention, 'mem_attn', epoch, val_loss, hub_id=self.hub_model_mem_attn)
                    model.memory_attention.model.stm.batched_memory(batch_size)

    def _save_final(self, model: Union[nn.Module, PyTorchModelHubMixin], component: str, hub_id: str = None):
        try:
            # Save final model
            if model.save_pretrained is not None:
                ckpt_path = os.path.join(
                    self.save_dir,
                    component,
                    'final_model'
                )
                path_exists = os.path.exists(ckpt_path)
                if not path_exists:
                    os.makedirs(ckpt_path)
                model.save_pretrained(save_directory=ckpt_path)
            else:
                comp_path = os.path.join(
                    self.save_dir,
                    component
                )
                path_exists = os.path.exists(comp_path)
                if not path_exists:
                    os.makedirs(comp_path)
                ckpt_path = os.path.join(comp_path, 'final_model.pt')
                torch.save(model.state_dict(), ckpt_path)
            print(f"Final model saved to {ckpt_path}")
        except Exception as e:
            print(f"Error saving final model: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()
        try:
            if self.push_to_hub and model.push_to_hub is not None and hub_id:
                model.push_to_hub(
                    repo_id=hub_id,
                    commit_message=self.final_commit_message or f'Final pre-trained model, after {self.finished_epochs} epochs',
                    token=self.hf_token,
                    private=self.private_repo,
                )
        except Exception as e:
            print(f"Error pushing final model: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

    def on_training_end(self, model: Union[nn.Module, PyTorchModelHubMixin]):
        if self.rank == 0:
            if isinstance(model, DistributedDataParallel):
                model = next(model.children())

            if 'decoder' in self.save_models:
                batch_size = model.decoder.model.stm.batch_size
                model.decoder.model.stm.single_memory()
                self._save_final(model.decoder, 'decoder', hub_id=self.hub_model_decoder)
                model.decoder.model.stm.batched_memory(batch_size)

            if 'encoder' in self.save_models:
                self._save_final(model.encoder, 'encoder', hub_id=self.hub_model_encoder)

            if 'head' in self.save_models:
                self._save_final(model.mlm_head, 'head', hub_id=self.hub_model_head)

            if 'mem_attn' in self.save_models:
                batch_size = model.memory_attention.model.stm.batch_size
                model.memory_attention.model.stm.single_memory()
                self._save_final(model.memory_attention, 'mem_attn', hub_id=self.hub_model_mem_attn)
                model.memory_attention.model.stm.batched_memory(batch_size)


class EarlyStoppageCallback(TrainerCallback):
    def __init__(self, num_plateau_epochs: int = 3) -> None:
        super().__init__()
        self.num_plateau_epochs = num_plateau_epochs
        self.best_loss = 9999.0
        self.best_loss_epoch = 0

    def on_validation_end(
            self,
            model: nn.Module,
            epoch: int,
            val_loss: float,
            val_metrics: dict
    ):
        if val_loss < self.best_loss:
            self.best_loss = val_loss
            self.best_loss_epoch = epoch
        elif epoch - self.best_loss_epoch >= self.num_plateau_epochs:
            return True
        return None


class MrlTrainerCallback:
    def on_epoch_start(self, actor: nn.Module, epoch: int, stage_epochs: int, global_epoch: int,
                       global_epochs: int, curriculum_config: dict) -> None:
        pass

    def on_epoch_end(self, actor: nn.Module, epoch: int, stage_epochs: int, policy_loss: float,
                     critic_loss: float, global_epoch: int, global_epochs: int) -> None:
        pass

    def on_episode_collected(self, actor: nn.Module, batch_idx: int, episode_trajectories: list[dict],
                             reward: float) -> None:
        pass

    def on_reward(
            self, actor: nn.Module, rewards: list[float], generated: dict[str, torch.Tensor],
            reference: dict[str, torch.Tensor], saved_data: dict[str, torch.Tensor],
            query: dict[str, torch.Tensor], eval_mode: bool
    ) -> None:
        pass

    def on_update_epoch_start(self, actor: nn.Module, critic: nn.Module, global_epoch: int, update_epoch: int) -> None:
        pass

    def on_batch_updated(self, actor: nn.Module, epoch: int, step: int, policy_loss: float) -> None:
        pass

    def on_critic_updated(self, actor: nn.Module, critic: nn.Module, epoch: int, step: int,
                          critic_loss: float) -> None:
        pass

    def on_update_epoch_end(self, actor: nn.Module, critic: nn.Module, global_epoch: int, update_epoch: int, policy_loss: float, critic_loss: float) -> None:
        pass

    def on_training_end(self, actor: nn.Module, critic: nn.Module, curriculum_config: dict) -> None:
        pass

    def on_eval_end(self, actor: nn.Module, critic: nn.Module, epoch: int, eval_mean_reward: float) -> Union[bool, None]:
        pass

    def on_eval_episode_end(self, actor: nn.Module, epoch: int, batch_idx: int, reward: float) -> None:
        pass


class MrlPrintCallback(MrlTrainerCallback):
    def __init__(
            self,
            update_steps_interval: int = 10,
            print_best_and_worst_generated: bool = False,
            tokenizer: Optional[Union[PreTrainedTokenizer, PreTrainedTokenizerFast]] = None,
            print_generated_only_on_eval: bool = False,
    ) -> None:
        super(MrlPrintCallback, self).__init__()
        self.update_steps_interval = update_steps_interval
        self.policy_losses = []
        self.critic_losses = []
        self.print_generated = print_best_and_worst_generated
        self.tokenizer = tokenizer
        self.print_generated_only_on_eval = print_generated_only_on_eval

    def on_epoch_start(self, actor: nn.Module, epoch: int, stage_epochs: int, curriculum_config: dict,
                       global_epoch: int, global_epochs: int) -> None:
        print(
            f'Starting epoch {epoch}/{stage_epochs - 1} (stage) | {global_epoch}/{global_epochs} (global) for {curriculum_config["steps"]} steps in {curriculum_config["strategy"]} strategy.')

    def on_epoch_end(self, actor: nn.Module, epoch: int, stage_epochs: int, policy_loss: float,
                     critic_loss: float, global_epoch: int, global_epochs: int) -> None:
        print(f'Finished epoch {epoch}/{stage_epochs - 1} (stage) | {global_epoch}/{global_epochs} (global)')
        print(f'Policy mean loss: {policy_loss} | Critic mean loss: {critic_loss}')

    def on_episode_collected(self, actor: nn.Module, batch_idx: int, episode_trajectories: list[dict],
                             reward: float) -> None:
        print(f'Collected {batch_idx} episode | mean reward {reward}')

    def _rewards_std(self, rewards: list[float]) -> float:
        mean_reward = sum(rewards) / len(rewards)
        diffs = [(r - mean_reward) ** 2 for r in rewards]
        return (sum(diffs) / len(diffs)) ** 0.5

    def _decode_sequence(self, sequence: dict[str, torch.Tensor]) -> str:
        seq_len = sequence['attention_mask'].sum().item()
        decoded = self.tokenizer.decode(sequence['input_ids'][:seq_len]).replace('Ċ', '\n').replace('Ġ', ' ')
        return decoded

    def _get_generated_text(
            self, generated: dict[str, torch.Tensor], reference: dict[str, torch.Tensor],
            saved_data: dict[str, torch.Tensor], query: dict[str, torch.Tensor]
    ):
        gen = self._decode_sequence(generated)
        ref = self._decode_sequence(reference)
        saved = self._decode_sequence(saved_data)
        query = self._decode_sequence(query)

        return gen, ref, saved, query

    def _get_sequence_on_index(self, batch: dict[str, torch.Tensor], index: int) -> dict[str, torch.Tensor]:
        return {
            'input_ids': batch['input_ids'][index],
            'attention_mask': batch['attention_mask'][index],
        }

    def _print_results(self, gen: str, ref: str, saved: str, query: str):
        print('Saved data (previous interaction):')
        print(saved)
        print('Query:')
        print(query)
        print('Reference:')
        print(ref)
        print('Generated:')
        print(gen)

    def print_generated_texts(
            self, rewards: list[float], generated: dict[str, torch.Tensor], reference: dict[str, torch.Tensor],
            saved_data: dict[str, torch.Tensor], query: dict[str, torch.Tensor]
    ):
        rewards_arr = np.array(rewards)
        best_index = int(np.argmax(rewards_arr))
        worst_index = int(np.argmin(rewards_arr))

        best_gen = self._get_sequence_on_index(generated, best_index)
        best_ref = self._get_sequence_on_index(reference, best_index)
        best_saved = self._get_sequence_on_index(saved_data, best_index)
        best_query = self._get_sequence_on_index(query, best_index)

        best_gen_txt, best_ref_txt, best_saved_txt, best_query_txt = self._get_generated_text(
            best_gen, best_ref, best_saved, best_query
        )

        print(f'Generated and reference results for the best reward: {max(rewards)}')
        self._print_results(best_gen_txt, best_ref_txt, best_saved_txt, best_query_txt)

        worst_gen = self._get_sequence_on_index(generated, worst_index)
        worst_ref = self._get_sequence_on_index(reference, worst_index)
        worst_saved = self._get_sequence_on_index(saved_data, worst_index)
        worst_query = self._get_sequence_on_index(query, worst_index)

        worst_gen_txt, worst_ref_txt, worst_saved_txt, worst_query_txt = self._get_generated_text(
            worst_gen, worst_ref, worst_saved, worst_query
        )

        print(f'Generated and reference results for the worst reward: {min(rewards)}')
        self._print_results(worst_gen_txt, worst_ref_txt, worst_saved_txt, worst_query_txt)

    def on_reward(
            self, actor: nn.Module, rewards: list[float], generated: dict[str, torch.Tensor],
            reference: dict[str, torch.Tensor], saved_data: dict[str, torch.Tensor],
            query: dict[str, torch.Tensor], eval_mode: bool
    ) -> None:
        print(f"{'Eval' if eval_mode else 'Train'} | Mean reward: {sum(rewards) / len(rewards)}, min: {min(rewards)}, max: {max(rewards)}, std: {self._rewards_std(rewards)} | All collected rewards: {rewards}")

        if self.print_generated and self.tokenizer is not None:
            if self.print_generated_only_on_eval and not eval_mode:
                return
            self.print_generated_texts(rewards, generated, reference, saved_data, query)


    def on_update_epoch_start(self, actor: nn.Module, critic: nn.Module, global_epoch: int, update_epoch: int) -> None:
        print(f'Epoch {global_epoch} | Starting update epoch {update_epoch}')

    def on_batch_updated(self, actor: nn.Module, epoch: int, step: int, policy_loss: float) -> None:
        if step != 0 and step % self.update_steps_interval == 0:
            loss = sum(self.policy_losses) / len(self.policy_losses)
            self.policy_losses = []
            print(f'Epoch {epoch} | Steps {step - self.update_steps_interval} - {step} - mean policy loss {loss} | current policy loss {policy_loss}')
        else:
            self.policy_losses.append(policy_loss)

    def on_critic_updated(self, actor: nn.Module, critic: nn.Module, epoch: int, step: int,
                          critic_loss: float) -> None:
        if step != 0 and step % self.update_steps_interval == 0:
            loss = sum(self.critic_losses) / len(self.critic_losses)
            self.critic_losses = []
            print(f'Epoch {epoch} | Steps {step - self.update_steps_interval} - {step} - mean critic loss {loss} | current critic loss {critic_loss}')
        else:
            self.critic_losses.append(critic_loss)

    def on_update_epoch_end(self, actor: nn.Module, critic: nn.Module, global_epoch: int, update_epoch: int, policy_loss: float, critic_loss: float) -> None:
        print(f'Epoch {global_epoch} | Update epoch {update_epoch} - mean policy loss {policy_loss} | mean critic loss {critic_loss}')

    def on_training_end(self, actor: nn.Module, critic: nn.Module, curriculum_config: dict) -> None:
        print(f'Finished training for {curriculum_config["steps"]} steps in {curriculum_config["strategy"]} strategy.')

    def on_eval_end(self, actor: nn.Module, critic: nn.Module, epoch: int, eval_mean_reward: float) -> None:
        print(f'Eval epoch {epoch} - mean reward {eval_mean_reward}')

    def on_eval_episode_end(self, actor: nn.Module, epoch: int, batch_idx: int, reward: float) -> None:
        print(f'Eval epoch {epoch} / Episode {batch_idx} - mean reward {reward}')


class MrlEarlyStoppageCallback(MrlTrainerCallback):
    def __init__(self, num_plateau_epochs: int = 2, threshold: Optional[float] = None) -> None:
        super().__init__()
        self.num_plateau_epochs = num_plateau_epochs
        self.best_reward = -9999.0
        self.best_reward_epoch = 0
        self.threshold = threshold

    def on_eval_end(self, _actor: nn.Module, _critic: nn.Module, epoch: int, eval_mean_reward: float) -> Union[bool, None]:
        if self.threshold is not None:
            if eval_mean_reward > self.threshold:
                return True

        if eval_mean_reward > self.best_reward:
            self.best_reward = eval_mean_reward
            self.best_reward_epoch = epoch
        elif epoch - self.best_reward_epoch >= self.num_plateau_epochs:
            return True
        return None

class MrlModelSaveCallback(MrlTrainerCallback):
    def __init__(
            self,
            save_dir: str,
            max_keep: int = 3,
            push_to_hub: bool = False,
            hub_model_decoder: str = None,
            hub_model_encoder: str = None,
            hub_model_memory_attention: str = None,
            hub_model_critic: str = None,
            private_repo: bool = False,
            hf_token: str = None,
            push_checkpoint_weights: bool = True,
            final_commit_message: str = None,
            display_exc_trace: bool = False,
            use_ddp: bool = False,
    ):
        self.save_dir = save_dir
        self.max_keep = max_keep
        self.best_reward = float('-inf')
        self.ckpt_paths = []
        self.push_to_hub = push_to_hub
        self.hub_model_decoder = hub_model_decoder
        self.hub_model_encoder = hub_model_encoder
        self.hub_model_memory_attention = hub_model_memory_attention
        self.hub_model_critic = hub_model_critic
        self.private_repo = private_repo
        self.hf_token = hf_token
        self.push_checkpoint_weights = push_checkpoint_weights
        self.final_commit_message = final_commit_message
        self.finished_epochs = 0
        self.display_exc_trace = display_exc_trace
        self.rank = int(os.environ['RANK']) if use_ddp else 0

    def _save_eval(self, model: Union[nn.Module, PyTorchModelHubMixin], component: str, epoch: int,
                         reward: float, hub_id: str = None):
        try:
            if model.save_pretrained is not None:
                ckpt_path = os.path.join(
                    self.save_dir,
                    component,
                    f'epoch_{epoch}_eval_reward_{reward:.4f}'
                )
                path_exists = os.path.exists(ckpt_path)
                if not path_exists:
                    os.makedirs(ckpt_path)
                model.save_pretrained(save_directory=ckpt_path)
            else:
                comp_path = os.path.join(
                    self.save_dir,
                    component
                )
                path_exists = os.path.exists(comp_path)
                if not path_exists:
                    os.makedirs(comp_path)
                ckpt_path = os.path.join(
                    comp_path,
                    f'epoch_{epoch}_eval_reward_{reward:.4f}.pt'
                )
                torch.save(model.state_dict(), ckpt_path)
            self.ckpt_paths.append(ckpt_path)

            # Keep only N best checkpoints
            if len(self.ckpt_paths) > self.max_keep:
                oldest_path = self.ckpt_paths.pop(0)
                if model.save_pretrained is not None:
                    shutil.rmtree(oldest_path)
                else:
                    os.remove(oldest_path)
        except Exception as e:
            print(f"Error saving epoch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

        try:
            if self.push_to_hub and self.push_checkpoint_weights and model.push_to_hub is not None and hub_id:
                model.push_to_hub(
                    repo_id=hub_id,
                    commit_message=f'Epoch {epoch} - Eval reward {reward:.4f}',
                    token=self.hf_token,
                    private=self.private_repo,
                )
        except Exception as e:
            print(f"Error pushing epoch checkpoint: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

    def on_eval_end(self, actor: nn.Module, critic: nn.Module, epoch: int, eval_mean_reward: float) -> None:
        if self.rank == 0:
            self.finished_epochs += 1
            if eval_mean_reward > self.best_reward:
                self.best_reward = eval_mean_reward
                if isinstance(actor, DistributedDataParallel):
                    actor = next(actor.children())

                batch_size = actor.decoder.model.stm.batch_size
                actor.decoder.model.stm.single_memory()

                self._save_eval(actor.encoder, 'encoder', epoch, eval_mean_reward, hub_id=self.hub_model_encoder)
                self._save_eval(actor.decoder, 'decoder', epoch, eval_mean_reward, hub_id=self.hub_model_decoder)
                self._save_eval(actor.memory_attention, 'memory_attention', epoch, eval_mean_reward, hub_id=self.hub_model_memory_attention)
                actor.decoder.model.stm.batched_memory(batch_size)

                if isinstance(critic, DistributedDataParallel):
                    critic = next(critic.children())
                self._save_eval(critic, 'critic', epoch, eval_mean_reward, hub_id=self.hub_model_critic)

    def _save_final(self, model: Union[nn.Module, PyTorchModelHubMixin], component: str, hub_id: str = None):
        try:
            # Save final model
            if model.save_pretrained is not None:
                ckpt_path = os.path.join(
                    self.save_dir,
                    component,
                    'final_model'
                )
                path_exists = os.path.exists(ckpt_path)
                if not path_exists:
                    os.makedirs(ckpt_path)
                model.save_pretrained(save_directory=ckpt_path)
            else:
                comp_path = os.path.join(
                    self.save_dir,
                    component
                )
                path_exists = os.path.exists(comp_path)
                if not path_exists:
                    os.makedirs(comp_path)
                ckpt_path = os.path.join(comp_path, 'final_model.pt')
                torch.save(model.state_dict(), ckpt_path)
            print(f"Final model saved to {ckpt_path}")
        except Exception as e:
            print(f"Error saving final model: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()
        try:
            if self.push_to_hub and model.push_to_hub is not None and hub_id:
                model.push_to_hub(
                    repo_id=hub_id,
                    commit_message=self.final_commit_message or f'Model after full curriculum stage, after {self.finished_epochs} epochs',
                    token=self.hf_token,
                    private=self.private_repo,
                )
        except Exception as e:
            print(f"Error pushing final model: {str(e)}")
            if self.display_exc_trace:
                traceback.print_exc()

    def on_training_end(self, actor: nn.Module, critic: nn.Module, curriculum_config: dict) -> None:
        if self.rank == 0:
            if isinstance(actor, DistributedDataParallel):
                actor = next(actor.children())

            batch_size = actor.decoder.model.stm.batch_size
            actor.decoder.model.stm.single_memory()

            self._save_final(actor.encoder, 'encoder', hub_id=self.hub_model_encoder)
            self._save_final(actor.decoder, 'decoder', hub_id=self.hub_model_decoder)
            self._save_final(actor.memory_attention, 'memory_attention', hub_id=self.hub_model_memory_attention)

            actor.decoder.model.stm.batched_memory(batch_size)

            if isinstance(critic, DistributedDataParallel):
                critic = next(critic.children())
            self._save_final(critic, 'critic', hub_id=self.hub_model_critic)

class MrlGeneratedTokensCallback(MrlTrainerCallback):
    def __init__(self, steps_log_interval: int = 10):
        self.total_tokens = 0
        self.steps_log_interval = steps_log_interval
        self.step = 0

    def on_reward(self, actor: nn.Module, rewards: list[float], generated: dict[str, torch.Tensor],
                  reference: dict[str, torch.Tensor], saved_data: dict[str, torch.Tensor], query: dict[str, torch.Tensor], eval_mode: bool) -> None:
        self.step += 1
        attention_mask = generated['attention_mask']
        batch_tokens = attention_mask.sum().item()
        self.total_tokens += batch_tokens
        if self.step != 0 and self.step % self.steps_log_interval == 0:
            print(f'Total processed tokens: {human_format(self.total_tokens)}')

    def on_training_end(self, actor: nn.Module, critic: nn.Module, curriculum_config: dict) -> None:
        print(f'Total training tokens: {human_format(self.total_tokens)}')

    def get_total_tokens(self):
        return self.total_tokens




import torch
import torch.nn as nn

class ShortTermMemory(nn.Module):
    """Short-term memory module for the Attention-based Memory System"""

    def __init__(self, num_layers: int, embed_dim: int, stm_size: int, init_type: str = 'normal',
                 is_trainable: bool = False, *args, **kwargs):
        super(ShortTermMemory, self).__init__(*args, **kwargs)
        self.num_layers = num_layers
        self.embed_dim = embed_dim
        self.stm_size = stm_size
        self.batch_size = 1 # setting 1 as initial batch size (it will be normally used in inference/pre-training. Bigger batches are for RL stages)
        self.is_trainable = is_trainable
        assert init_type in ['normal', 'standard', 'uniform', 'ones', 'zeros'], \
            'STM init type must be one of "normal", "standard", "uniform", "ones", "zeros"'
        self.init_type = init_type
        stm = self._init_tensor()
        if self.is_trainable:
            self.memory = nn.Parameter(stm)
        else:
            self.register_buffer('memory', stm)

    def _init_tensor(self, init_type: str = None):
        init_type = init_type or self.init_type
        stm_shape = (self.num_layers, self.batch_size, self.stm_size, self.embed_dim)
        if init_type == 'normal':
            return torch.normal(0, 0.02, stm_shape)
        elif init_type == 'standard':
            return torch.normal(0, 1, stm_shape)
        elif init_type == 'uniform':
            return torch.rand(*stm_shape) * 0.02
        elif init_type == 'ones':
            return torch.ones(*stm_shape)
        else:
            return torch.zeros(*stm_shape)

    def forward(self, layer: int) -> torch.Tensor:
        return self.memory[layer]

    def update_layer(self, layer: int, new_stm: torch.Tensor):
        self.memory = self.memory.clone()
        self.memory[layer] = new_stm

    def update_all(self, new_stm: torch.Tensor):
        self.memory = new_stm

    def make_trainable(self):
        if not self.is_trainable:
            self.is_trainable = True
            initial_stm = self.memory.clone()
            delattr(self, 'memory')
            self.memory = nn.Parameter(initial_stm)

    def freeze(self):
        if self.is_trainable:
            self.requires_grad_(False)
            trained_stm = self.memory.clone()
            delattr(self, 'memory')
            self.register_buffer('memory', trained_stm)

    def reset(self, init_type: str = None):
        self.memory = self._init_tensor(init_type).to(self.memory.device)

    def clone_detach_reset(self):
        self.memory = self.memory.detach().clone()

    def resize(self, new_stm_size: int, init_type: str = None):
        self.stm_size = new_stm_size
        delattr(self, 'memory')
        self.register_buffer('memory', self._init_tensor(init_type))

    def batched_memory(self, batch_size: int, init_type: str = None):
        if init_type is not None:
            assert init_type in ['normal', 'standard', 'uniform', 'ones', 'zeros'], \
                'STM init type must be one of "normal", "standard", "uniform", "ones", "zeros"'
            self.init_type = init_type
        device = self.memory.device
        self.batch_size = batch_size
        delattr(self, 'memory')
        self.register_buffer('memory', self._init_tensor().to(device))

    def single_memory(self, init_type: str = None, use_mean_from_batch: bool = False):
        if init_type is not None:
            assert init_type in ['normal', 'standard', 'uniform', 'ones', 'zeros'], \
                'STM init type must be one of "normal", "standard", "uniform", "ones", "zeros"'
            self.init_type = init_type
        device = self.memory.device
        self.batch_size = 1
        if use_mean_from_batch:
            batch_mean = self.memory.mean(dim=(1, 2, 3), keepdim=True)
            delattr(self, 'memory')
            self.register_buffer('memory', batch_mean)
        else:
            delattr(self, 'memory')
            self.register_buffer('memory', self._init_tensor().to(device))



import torch
from torch.optim.lr_scheduler import LambdaLR, CosineAnnealingLR
import math

def get_transformer_lr_scheduler(
    optimizer: torch.optim.Optimizer,
    num_training_steps: int,
    warmup_steps: int = 0
):
    if warmup_steps > 0:
        # Warmup + cosine decay
        def lr_lambda(current_step):
            if current_step < warmup_steps:
                return float(current_step) / max(1, warmup_steps)
            remaining = max(0, current_step - warmup_steps)
            return 0.5 * (1.0 + math.cos(math.pi * remaining / (num_training_steps - warmup_steps)))
        return LambdaLR(optimizer, lr_lambda)
    else:
        return CosineAnnealingLR(optimizer, T_max=num_training_steps)

def calculate_steps(
        dataset_size: int,
        epochs: int,
        batch_size: int,
        warmup_ratio: float = 0.0,
        num_workers: int = 1,
        gradient_accumulation_steps: int = 1,
        verbose: bool = True,
):
    steps_per_epoch = int((dataset_size / batch_size - 1) // num_workers)
    total_steps = int((epochs * steps_per_epoch) / gradient_accumulation_steps)
    warmup_steps = int(warmup_ratio * total_steps)
    if verbose:
        print(f'Total steps: {total_steps}')
        print(f'Warmup steps: {warmup_steps}')
        print(f'Total steps per epoch: {steps_per_epoch}')
    return { 'total': total_steps, 'warmup': warmup_steps, 'epoch': steps_per_epoch}

def calculate_steps_for_smst(
        dataset_size: int,
        epochs: int,
        curriculum_steps: int,
        batch_size: int,
        warmup_ratio: float = 0.0,
        num_workers: int = 1,
        gradient_accumulation_steps: int = 1,
        verbose: bool = True,
):
    batches_per_epoch = int(((dataset_size / batch_size - 1) // num_workers))
    inner_steps_per_epoch = int(batches_per_epoch * curriculum_steps)

    total_steps = int((epochs * inner_steps_per_epoch) / gradient_accumulation_steps)
    warmup_steps = int(warmup_ratio * total_steps)
    if verbose:
        print(f'Total steps: {total_steps}')
        print(f'Warmup steps: {warmup_steps}')
        print(f'Total batches per epoch: {batches_per_epoch}')
        print(f'Total steps per epoch: {inner_steps_per_epoch}')
    return { 'total': total_steps, 'warmup': warmup_steps, 'epoch': inner_steps_per_epoch, 'batch': batches_per_epoch }




import torch
from typing import Optional, Literal


def tensor_sentence_bleu(
        candidates: torch.Tensor,
        references: torch.Tensor,
        pad_token_id: int = 0,
        max_n: int = 4,
        weights: Optional[torch.Tensor] = None,
        smoothing_method: Literal['none', 'floor', 'add-k', 'exp'] = 'exp',
        smoothing_epsilon: float = 0.1,
        smoothing_k: int = 1
) -> torch.Tensor:
    """
    Computes the BLEU score for each sentence in a batch.

    Args:
        candidates (torch.Tensor): Tensor token ids for candidate sentences. Shape: (batch_size, candidate_len).
        references (torch.Tensor): Tensor token ids for reference sentences. Shape: (batch_size, num_refs, ref_len).
        pad_token_id (int): Ignored padding token id (default 0).
        max_n (int): Maximum order of n-grams to include (default 4).
        weights (torch.Tensor): Weights for each row of n-grams. Tensor of dimension (max_n,).
        smoothing_method (str): Smoothing method. One from: ('none', 'floor', 'add-k', 'exp').
        smoothing_epsilon (float): Epsilon value for 'floor' method (default 0.1).
        smoothing_k (int): K value for 'add-k' method (default 1).

    Returns:
        torch.Tensor: A tensor of shape (batch_size,) containing the BLEU score for each sentence.
    """
    if weights is None:
        weights = torch.ones(max_n, device=candidates.device) / max_n

    max_n = weights.size(0)
    batch_size = candidates.size(0)
    num_refs = references.size(1)

    candidate_mask = (candidates != pad_token_id)
    reference_mask = (references != pad_token_id)

    candidate_lengths = candidate_mask.sum(dim=1)
    reference_lengths = reference_mask.sum(dim=2)

    # Track numerators and denominators per sentence
    numerators = torch.zeros(batch_size, max_n, device=candidates.device)
    denominators = torch.zeros(batch_size, max_n, device=candidates.device)

    for n in range(1, max_n + 1):
        # Skip sentences that are too short for this n-gram order
        valid_for_n = (candidate_lengths >= n)
        if not torch.any(valid_for_n):
            continue

        # We only need to process sentences long enough for the current n-gram
        candidates_n = candidates[valid_for_n]
        references_n = references[valid_for_n]
        batch_size_n = candidates_n.size(0)

        candidate_ngrams = candidates_n.unfold(1, n, 1)
        candidate_mask_n = candidate_mask[valid_for_n]
        valid_candidate_mask = candidate_mask_n.unfold(1, n, 1).all(dim=2)
        candidate_ngrams_flat = candidate_ngrams[valid_candidate_mask]

        # Denominators are the number of n-grams in each candidate sentence
        denominators[valid_for_n, n - 1] = valid_candidate_mask.sum(dim=1)

        if candidate_ngrams_flat.size(0) == 0:
            continue

        reference_ngrams_list = []
        for i in range(num_refs):
            ref_i = references_n[:, i, :]
            ref_mask_i = reference_mask[valid_for_n, i, :]
            if torch.any(ref_mask_i.sum(dim=1) < n):
                # This part is complex to handle per-sentence, so we simplify
                # by unfolding and masking, which handles empty cases correctly.
                pass
            ref_ngrams_i = ref_i.unfold(1, n, 1)
            valid_ref_mask_i = ref_mask_i.unfold(1, n, 1).all(dim=2)
            reference_ngrams_list.append(ref_ngrams_i[valid_ref_mask_i])

        reference_ngrams_flat = torch.cat(reference_ngrams_list, dim=0)
        all_ngrams = torch.cat([candidate_ngrams_flat, reference_ngrams_flat], dim=0)

        if all_ngrams.size(0) == 0:
            continue

        unique_ngrams, inverse_indices = torch.unique(all_ngrams, dim=0, return_inverse=True)
        num_unique = unique_ngrams.size(0)

        cand_inv_indices = inverse_indices[:candidate_ngrams_flat.size(0)]
        ref_inv_indices = inverse_indices[candidate_ngrams_flat.size(0):]

        cand_batch_indices = torch.arange(batch_size_n, device=candidates.device).unsqueeze(1).expand_as(
            valid_candidate_mask)
        cand_batch_indices_flat = cand_batch_indices[valid_candidate_mask]
        cand_offsets = cand_batch_indices_flat * num_unique
        cand_bincount_indices = cand_inv_indices + cand_offsets
        candidate_counts = torch.bincount(
            cand_bincount_indices,
            minlength=batch_size_n * num_unique
        ).reshape(batch_size_n, num_unique)

        ref_max_counts = torch.zeros_like(candidate_counts)
        start_idx = 0
        for i in range(num_refs):
            ref_mask_i = reference_mask[valid_for_n, i, :]
            valid_ref_mask_i = ref_mask_i.unfold(1, n, 1).all(dim=2)
            num_ref_ngrams_i = valid_ref_mask_i.sum()

            if num_ref_ngrams_i == 0:
                continue

            ref_inv_indices_i = ref_inv_indices[start_idx: start_idx + num_ref_ngrams_i]
            start_idx += num_ref_ngrams_i

            ref_batch_indices = torch.arange(batch_size_n, device=candidates.device).unsqueeze(1).expand_as(
                valid_ref_mask_i)
            ref_batch_indices_flat = ref_batch_indices[valid_ref_mask_i]
            ref_offsets = ref_batch_indices_flat * num_unique
            ref_bincount_indices = ref_inv_indices_i + ref_offsets
            ref_counts_i = torch.bincount(
                ref_bincount_indices,
                minlength=batch_size_n * num_unique
            ).reshape(batch_size_n, num_unique)
            ref_max_counts = torch.maximum(ref_max_counts, ref_counts_i)

        clipped_counts = torch.minimum(candidate_counts, ref_max_counts)

        # Numerators are the sum of clipped counts for each sentence
        numerators[valid_for_n, n - 1] = clipped_counts.sum(dim=1)

    # All subsequent calculations are now element-wise for the batch
    precisions = torch.zeros(batch_size, max_n, device=candidates.device)
    if smoothing_method == 'floor':
        safe_numerators = torch.where(numerators == 0, smoothing_epsilon, numerators)
        safe_denominators = torch.where(denominators == 0, 1.0, denominators)
        precisions = safe_numerators / safe_denominators
    elif smoothing_method == 'add-k':
        precisions = (numerators + smoothing_k) / (denominators + smoothing_k)
    # The 'exp' smoothing method is tricky for per-sentence scores and less standard.
    # A simple approach is to fall back to a small value.
    else:  # 'none' or a simplified 'exp'
        mask = denominators > 0
        precisions[mask] = numerators[mask] / denominators[mask]
        # For 'exp' smoothing or to avoid log(0)
        if smoothing_method == 'exp':
            precisions[precisions == 0] = 1e-12

    closest_ref_lengths_indices = torch.abs(reference_lengths - candidate_lengths.unsqueeze(1)).argmin(dim=1)
    closest_ref_lengths = reference_lengths.gather(1, closest_ref_lengths_indices.unsqueeze(1)).squeeze(1)

    # Calculate brevity penalty per sentence
    brevity_penalty = torch.ones(batch_size, device=candidates.device)
    mask = candidate_lengths < closest_ref_lengths
    brevity_penalty[mask] = torch.exp(1 - closest_ref_lengths[mask] / candidate_lengths[mask])

    # Handle zero-length candidates
    brevity_penalty[candidate_lengths == 0] = 0.0

    # Final score calculation per sentence
    log_precisions = torch.log(precisions + 1e-12)  # Add epsilon to avoid log(0)
    weighted_log_precisions = weights.to(candidates.device) * log_precisions
    geo_mean = torch.exp(torch.sum(weighted_log_precisions, dim=1))

    score = brevity_penalty * geo_mean

    return score

def tensor_corpus_bleu(
    candidates: torch.Tensor,
    references: torch.Tensor,
    pad_token_id: int = 0,
    max_n: int = 4,
    weights: Optional[torch.Tensor] = None,
    smoothing_method: Literal['none', 'floor', 'add-k', 'exp'] = 'exp',
    smoothing_epsilon: float = 0.1,
    smoothing_k: int = 1
) -> torch.Tensor:
    """
    Computes the BLEU score for a batch of sequences in a fully vectorized and memory-efficient manner.
    This version avoids the out-of-memory problem by counting unique n-grams in a compact space instead of in a massive hash space.

    Args:
        candidates (torch.Tensor): Tensor token ids for candidate sentences. Shape: (batch_size, candidate_len).
        references (torch.Tensor): Tensor token ids for reference sentences. Shape: (batch_size, num_refs, ref_len).
        pad_token_id (int): Ignored padding token id (default 0).
        max_n (int): Maximum order of n-grams to include (default 4).
        weights (torch.Tensor): Weights for each row of n-grams. Tensor of dimension (max_n,).
        smoothing_method (str): Smoothing method. One from: ('none', 'floor', 'add-k', 'exp').
        smoothing_epsilon (float): Epsilon value for 'floor' method (default 0.1).
        smoothing_k (int): K value for 'add-k' method (default 1).

    Returns:
        torch.Tensor: A scalar tensor containing the BLEU score for the entire batch.
    """
    if weights is None:
        weights = torch.ones(max_n, device=candidates.device) / max_n

    max_n = weights.size(0)

    batch_size = candidates.size(0)
    num_refs = references.size(1)

    candidate_mask = (candidates!= pad_token_id)
    reference_mask = (references!= pad_token_id)

    candidate_lengths = candidate_mask.sum(dim=1)
    reference_lengths = reference_mask.sum(dim=2)

    numerators = torch.zeros(max_n, device=candidates.device)
    denominators = torch.zeros(max_n, device=candidates.device)

    for n in range(1, max_n + 1):
        if torch.any(candidate_lengths < n):
            continue

        candidate_ngrams = candidates.unfold(1, n, 1)
        valid_candidate_mask = candidate_mask.unfold(1, n, 1).all(dim=2)
        candidate_ngrams_flat = candidate_ngrams[valid_candidate_mask]

        reference_ngrams_list =[]
        for i in range(num_refs):
            ref_i = references[:, i, :]
            ref_mask_i = reference_mask[:, i, :]
            if torch.any(ref_mask_i.sum(dim=1) < n):
                reference_ngrams_list.append(torch.tensor([], dtype=torch.long, device=candidates.device).reshape(0, n))
                continue
            ref_ngrams_i = ref_i.unfold(1, n, 1)
            valid_ref_mask_i = ref_mask_i.unfold(1, n, 1).all(dim=2)
            reference_ngrams_list.append(ref_ngrams_i[valid_ref_mask_i])

        if candidate_ngrams_flat.size(0) == 0:
            continue

        reference_ngrams_flat = torch.cat(reference_ngrams_list, dim=0)

        all_ngrams = torch.cat([candidate_ngrams_flat, reference_ngrams_flat], dim=0)

        if all_ngrams.size(0) == 0:
            denominators[n-1] = candidate_ngrams_flat.size(0)
            continue

        unique_ngrams, inverse_indices = torch.unique(all_ngrams, dim=0, return_inverse=True)

        num_unique = unique_ngrams.size(0)

        cand_inv_indices = inverse_indices[:candidate_ngrams_flat.size(0)]
        ref_inv_indices = inverse_indices[candidate_ngrams_flat.size(0):]

        cand_batch_indices = torch.arange(batch_size, device=candidates.device).unsqueeze(1).expand_as(valid_candidate_mask)
        cand_batch_indices_flat = cand_batch_indices[valid_candidate_mask]

        cand_offsets = cand_batch_indices_flat * num_unique
        cand_bincount_indices = cand_inv_indices + cand_offsets

        candidate_counts = torch.bincount(
            cand_bincount_indices,
            minlength=batch_size * num_unique
        ).reshape(batch_size, num_unique)

        ref_max_counts = torch.zeros_like(candidate_counts)

        start_idx = 0
        for i in range(num_refs):
            ref_i = references[:, i, :]
            ref_mask_i = reference_mask[:, i, :]
            if torch.any(ref_mask_i.sum(dim=1) < n):
                continue

            valid_ref_mask_i = ref_mask_i.unfold(1, n, 1).all(dim=2)
            num_ref_ngrams_i = valid_ref_mask_i.sum()

            if num_ref_ngrams_i == 0:
                continue

            ref_inv_indices_i = ref_inv_indices[start_idx : start_idx + num_ref_ngrams_i]
            start_idx += num_ref_ngrams_i

            ref_batch_indices = torch.arange(batch_size, device=candidates.device).unsqueeze(1).expand_as(valid_ref_mask_i)
            ref_batch_indices_flat = ref_batch_indices[valid_ref_mask_i]

            ref_offsets = ref_batch_indices_flat * num_unique
            ref_bincount_indices = ref_inv_indices_i + ref_offsets

            ref_counts_i = torch.bincount(
                ref_bincount_indices,
                minlength=batch_size * num_unique
            ).reshape(batch_size, num_unique)

            ref_max_counts = torch.maximum(ref_max_counts, ref_counts_i)

        clipped_counts = torch.minimum(candidate_counts, ref_max_counts)

        numerators[n-1] = clipped_counts.sum()
        denominators[n-1] = candidate_ngrams_flat.size(0)

    precisions = torch.zeros(max_n, device=candidates.device)
    if smoothing_method == 'floor':
        safe_numerators = torch.where(numerators == 0, smoothing_epsilon, numerators)
        safe_denominators = torch.where(denominators == 0, 1.0, denominators)
        precisions = safe_numerators / safe_denominators
    elif smoothing_method == 'add-k':
        precisions = (numerators + smoothing_k) / (denominators + smoothing_k)
    elif smoothing_method == 'exp':
        for i in range(max_n):
            if denominators[i] > 0:
                precisions[i] = numerators[i] / denominators[i]
            else:
                precisions[i] = torch.tensor(0.0, device=candidates.device)
        for i in range(max_n):
            if precisions[i] == 0:
                if i == 0:
                    precisions[i] = torch.tensor(1e-12, device=candidates.device)
                else:
                    precisions[i] = precisions[i-1] / 2.0
    else: # 'none'
        if torch.any(denominators == 0):
            return torch.tensor(0.0, device=candidates.device)
        precisions = numerators / denominators

    closest_ref_lengths_indices = torch.abs(reference_lengths - candidate_lengths.unsqueeze(1)).argmin(dim=1)
    closest_ref_lengths = reference_lengths.gather(1, closest_ref_lengths_indices.unsqueeze(1)).squeeze(1)

    total_candidate_len = candidate_lengths.sum()
    total_closest_ref_len = closest_ref_lengths.sum()

    if total_candidate_len == 0:
        return torch.tensor(0.0, device=candidates.device)

    brevity_penalty = torch.tensor(1.0, device=candidates.device)
    if total_candidate_len < total_closest_ref_len:
        brevity_penalty = torch.exp(1 - total_closest_ref_len / total_candidate_len)

    log_precisions = torch.log(precisions + (1e-12 if smoothing_method == 'none' else 0)).to(candidates.device)
    score = brevity_penalty * torch.exp(torch.sum(weights.to(candidates.device) * log_precisions))

    return score